var ogma = new Ogma({
  container: 'graph-container',
  options: { backgroundColor: null }
});

// node types
var PORT = 'port';
var SERVER = 'server';
var IP = 'ip';

// edge types
var LISTENS_TO = 'listens_to';
var CONNECTS_TO = 'connects_to';

var ANIMATION_DURATION = 500;

var appState = {
  edgeGrouping: undefined,
  selectedPorts: {},
  timeFilter: [-Infinity, Infinity]
};

// transformations
var transformations = {
  edgeGrouping: null,
  portFilter: null,
  timeFilter: null
};

function initTransformations() {
  // port filtering policy
  transformations.portFilter = ogma.transformations.addNodeFilter({
    criteria: function(node) {
      if (node.getData('type') === PORT) {
        return appState.selectedPorts[node.getData('port')];
      } else {
        return true;
      }
    },
    enabled: true
  });

  // order matters, this has to go before the edge grouping
  transformations.timeFilter = ogma.transformations.addEdgeFilter({
    criteria: function(edge) {
      if (edge.getData('type') === CONNECTS_TO) {
        var date = edge.getData('date') || appState.timeFilter[0];
        return appState.timeFilter[0] <= date && date <= appState.timeFilter[1];
      }
      return true;
    }
  });

  // Edge grouping policy:
  //
  // group all edges between an “IP” and a “PORT” node by (protocol + service)
  // Grouped edge data:
  //   - protocol
  //   - service
  //   - total query bytes
  //   - total response bytes
  //   - group size (number of edges in the group)
  transformations.edgeGrouping = ogma.transformations.addEdgeGrouping({
    selector: function(edge) {
      return !edge.isExcluded() && edge.getData('type') === CONNECTS_TO;
    },
    groupIdFunction: function(e) {
      return e.getData('protocol') + '/' + e.getData('service');
    },
    generator: function(edges) {
      var e = edges.get(0);
      var protocol = e.getData('protocol');
      var service = e.getData('service');
      var query_bytes = 0,
        response_bytes = 0;
      edges.getData().forEach(function(edgeData) {
        query_bytes += edgeData.query_bytes || 0;
        response_bytes += edgeData.response_bytes || 0;
      });

      return {
        data: {
          type: CONNECTS_TO,
          query_bytes: query_bytes,
          response_bytes: response_bytes,
          group_size: edges.size,
          protocol: protocol,
          service: service
        }
      };
    },
    enabled: appState.edgeGrouping
  });
}

function layout() {
  return ogma.layouts.forceLink({
    iterationsPerRender: 200,
    barnesHutOptimize: true,
    barnesHutTheta: 0.3,
    gravity: 0.3
  });
}

// format traffic values
function formatPutThrough(bytes) {
  bytes = +bytes;
  return bytes > 1024 * 1024
    ? Math.round(bytes / (1024 * 1024)) + 'M'
    : bytes > 1024
    ? Math.round(bytes / 1024) + 'KB'
    : bytes + 'B';
}

// Style the graph based on the properties of nodes and edges
function addStyles() {
  ogma.styles.addNodeRule(
    function(node) {
      return node.getData('type') === SERVER;
    },
    {
      radius: 10,
      icon: {
        font: 'Font Awesome 5 Free',
        content: '\uF233',
        color: '#ffffff',
        style: 'bold',
        scale: 0.5
      }
    }
  );

  // node styles
  ogma.styles.addNodeRule(
    function(n) {
      return n.getData('type') === IP;
    },
    {
      color: '#222277',
      radius: 20,
      icon: {
        font: 'Font Awesome 5 Free',
        content: '\uF007',
        style: 'bold',
        color: '#ffffff',
        scale: 0.4
      },
      text: {
        content: function(node) {
          return node.getData('ip');
        },
        backgroundColor: 'white'
      },
      badges: {
        // number of outgoing “CONNECTS_TO” edges (taking filters into account)
        topRight: {
          text: {
            content: function(node) {
              return node
                .getAdjacentEdges({ direction: 'out ' })
                .filter(function(edge) {
                  return edge.getData('type') === CONNECTS_TO;
                }).size;
            },
            scale: 0.2
          },
          stroke: { width: 1 }
        },
        // received bytes (sum of all “response_bytes” from outgoing
        // “CONNECTS_TO” edges, taking filters into account)
        bottomLeft: {
          text: {
            content: function(node) {
              return formatPutThrough(
                node
                  .getAdjacentEdges({ direction: 'out' })
                  .filter(function(edge) {
                    return edge.getData('type') === CONNECTS_TO;
                  })
                  .reduce(function(total, edge) {
                    return total + edge.getData('response_bytes');
                  }, 0)
              );
            },
            scale: 0.2
          },
          stroke: { width: 1 }
        },

        // sent bytes (sum of all “query_bytes” from outgoing “CONNECTS_TO” edges,
        // taking filters into account)
        bottomRight: {
          text: {
            content: function(node) {
              return formatPutThrough(
                node
                  .getAdjacentEdges({ direction: 'out' })
                  .filter(function(edge) {
                    return edge.getData('type') === CONNECTS_TO;
                  })
                  .reduce(function(total, edge) {
                    return total + edge.getData('query_bytes');
                  }, 0)
              );
            },
            scale: 0.2
          },
          stroke: { width: 1 }
        }
      }
    }
  );

  ogma.styles.addNodeRule(
    function(n) {
      return n.getData('type') === PORT;
    },
    {
      color: 'orange',
      text: {
        content: function(node) {
          return ':' + node.getData('port');
        },
        position: 'center'
      },
      badges: {
        // number of incoming “CONNECTS_TO” edges (taking filters into account)
        topLeft: {
          text: function(node) {
            return node
              .getAdjacentEdges({ direction: 'in' })
              .filter(function(edge) {
                return edge.getData('type') === CONNECTS_TO;
              }).size;
          },
          stroke: { width: 1 }
        },

        // received bytes (sum of all “query_bytes” from incoming “CONNECTS_TO”
        // edges, taking filters into account)
        bottomLeft: {
          text: function(node) {
            return formatPutThrough(
              node
                .getAdjacentEdges({ direction: 'in' })
                .filter(function(edge) {
                  return edge.getData('type') === CONNECTS_TO;
                })
                .reduce(function(total, edge) {
                  return total + edge.getData('query_bytes');
                }, 0)
            );
          },
          stroke: { width: 1 }
        },

        // sent bytes (sum of all “response_bytes” from incoming “CONNECTS_TO”
        // edges, taking filters into account)
        bottomRight: {
          text: function(node) {
            return formatPutThrough(
              node
                .getAdjacentEdges({ direction: 'in' })
                .filter(function(edge) {
                  return edge.getData('type') === CONNECTS_TO;
                })
                .reduce(function(total, edge) {
                  return total + edge.getData('response_bytes');
                }, 0)
            );
          },
          stroke: { width: 1 }
        }
      }
    }
  );

  ogma.styles.addEdgeRule({
    color: function(edge) {
      return edge.getData('type') === CONNECTS_TO ? '#404050' : '#405040';
    },
    width: function(edge) {
      return 2 + Math.sqrt(edge.getData('group_size'));
    },
    shape: 'arrow',
    text: {
      content: function(edge) {
        if (edge.getData('type') === CONNECTS_TO) {
          var groupSize = edge.getData('group_size');
          return CONNECTS_TO + isNaN(groupSize) ? '' : ', ' + groupSize;
        } else {
          return LISTENS_TO;
        }
      }
    }
  });
}

function updateState() {
  // port filters
  var ports = document.querySelectorAll('#ports input');
  Array.prototype.forEach.call(ports, function(portInput) {
    var port = portInput.getAttribute('data-port-id');
    appState.selectedPorts[port] = portInput.checked;
  });

  // edge grouping
  appState.edgeGrouping = document.getElementById('group-edges').checked;

  // time filter
  var promise = appState.edgeGrouping
    ? transformations.edgeGrouping.enable(ANIMATION_DURATION)
    : transformations.edgeGrouping.disable(ANIMATION_DURATION);

  return promise
    .then(function() {
      return transformations.portFilter.refresh(ANIMATION_DURATION);
    })
    .then(function() {
      return ogma.transformations.afterNextUpdate();
    });
}

function initUI(graph) {
  createPortSelector();
  appState.timeFilter = ogma
    .getEdges()
    .filter(function(edge) {
      return edge.getData('type') === CONNECTS_TO;
    })
    .getData('date')
    .reduce(
      function(acc, date) {
        acc[0] = Math.min(acc[0], date);
        acc[1] = Math.max(acc[1], date);
        return acc;
      },
      [Infinity, -Infinity]
    );
  new TimeFilter(document.getElementById('timeline'), function() {
    if (transformations.timeFilter) {
      transformations.timeFilter.refresh();
    }
  }).update();
  document.getElementById('filters').addEventListener('change', updateState);
}

function createPortSelector(graph) {
  var portValues = {};
  appState.selectedPorts = {};
  ogma
    .getNodes()
    .filter(function(node) {
      return node.getData('type') === PORT;
    })
    .forEach(function(node) {
      var ref = node.getData();
      var port = ref.port;
      portValues[port] = portValues[port] || [];
      portValues[port].push(node.getId());
      appState.selectedPorts[port] = true;
    });

  var container = document.getElementById('ports');
  appState.ports = portValues;
  container.innerHTML = Object.keys(portValues)
    .sort(function(a, b) {
      return portValues[b].length - portValues[a].length;
    })
    .map(function(port) {
      return (
        '<input type="checkbox" data-port-id="' +
        port +
        '" ' +
        'checked id="port-' +
        port +
        '" name="port-' +
        port +
        '">' +
        '<label for="port-' +
        port +
        '" title="' +
        portValues[port].length +
        ' connections"> ' +
        port +
        '</label>'
      );
    })
    .join('');
}

loadAndParseCSV('logs-merged_sample.csv')
  .then(function(records) {
    return recordsToGraph(records.data);
  })
  .then(function(graph) {
    return ogma.setGraph(graph);
  })
  .then(function(graph) {
    return initUI(graph);
  })
  .then(initTransformations)
  .then(function() {
    return updateState();
  })
  .then(function() {
    return addStyles();
  })
  .then(initUI)
  .then(function() {
    return layout();
  })
  .then(function() {
    return ogma.view.locateGraph({ padding: 20 });
  });

function recordsToGraph(records) {
  // deduplication lookup hashes
  var portsMap = {};
  var ipsMap = {};
  var serversMap = {};
  var listensMap = {};

  var nodes = [],
    edges = [];

  records.forEach(function(record) {
    var timestamp = record[0];
    var src_ip = record[1];
    var dest_ip = record[2];
    var dest_port = record[3];
    var protocol = record[4];
    var service = record[5];
    var duration = record[6];
    var out_bytes = record[7];
    var in_bytes = record[8];
    var status = record[9];

    // Port node
    var portId = dest_ip + ':' + dest_port;
    var port = portsMap[portId];
    if (!port) {
      port = portsMap[portId] = {
        id: portId,
        data: {
          ip: dest_ip,
          port: dest_port,
          type: PORT
        }
      };
      nodes.push(port);
    }

    // Server node
    var serverId = dest_ip;
    var server = serversMap[serverId];
    if (!server) {
      server = serversMap[serverId] = {
        id: serverId,
        data: {
          ip: dest_ip,
          type: SERVER
        }
      };
      nodes.push(server);
    }

    // "Listens to" edge
    var listen = listensMap[portId];
    if (!listen) {
      listen = listensMap[portId] = {
        id: portId,
        source: serverId,
        target: portId,
        data: {
          type: LISTENS_TO
        }
      };
      edges.push(listen);
    }

    // IP node
    var ipId = src_ip;
    var ip = ipsMap[ipId];
    if (!ip) {
      ip = ipsMap[ipId] = {
        id: ipId,
        data: {
          ip: src_ip,
          type: IP
        }
      };
      nodes.push(ip);
    }

    // Connection edge
    var connectionId =
      src_ip +
      '-' +
      dest_ip +
      ':' +
      dest_port +
      '@' +
      Math.round(timestamp * 1000);
    var connection = {
      id: connectionId,
      source: ipId,
      target: portId,
      data: {
        type: CONNECTS_TO,
        date: Math.round(timestamp * 1000),
        query_bytes: isNaN(out_bytes) ? 0 : parseInt(out_bytes),
        response_bytes: isNaN(in_bytes) ? 0 : parseInt(in_bytes),
        protocol: protocol,
        service: service
      }
    };
    edges.push(connection);
  });

  return { nodes: nodes, edges: edges };
}

// load data
function loadAndParseCSV(url) {
  return new Promise(function(complete, error) {
    Papa.parse(url, {
      download: true,
      complete: complete,
      error: error
    });
  });
}
